/*
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <bluetooth/services/nus.h>
#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/direction.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/drivers/uart.h>
#include <zephyr/kernel.h>
#include <zephyr/logging/log.h>
#include <zephyr/sys/printk.h>
#include "bt_ad.h"
#include "bt_nus.h"
#include "error.h"

#define BLE_RX_BLOCK_SIZE              (CONFIG_BT_L2CAP_TX_MTU - 3)
#define BLE_RX_BUF_COUNT               4
#define BLE_SLAB_ALIGNMENT             4
#define CONFIG_BT_NUS_SECURITY_ENABLED 1

LOG_MODULE_REGISTER(bt);
K_MEM_SLAB_DEFINE(ble_rx_slab, BLE_RX_BLOCK_SIZE, BLE_RX_BUF_COUNT, BLE_SLAB_ALIGNMENT);

unsigned int _gPasskey = 123456;
static struct bt_conn* _gCurrentConn;
static struct bt_conn* auth_conn;

static void connected(struct bt_conn* conn, uint8_t err) {
    printk("connected() e\n");
    char addr[BT_ADDR_LE_STR_LEN];

    if (err) {
        printk("Connection failed (err %u)", err);
        return;
    }

    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
    printk("Connected %s\n", addr);

    _gCurrentConn = bt_conn_ref(conn);
    printk("_gCurrentConn=%p\n", _gCurrentConn);
    //printk("bt_nus_get_mtu(%p)=%u\n", _gCurrentConn, bt_nus_get_mtu(_gCurrentConn));

    const int ret = bt_conn_set_security(conn, BT_SECURITY_L2);
    if (ret != 0) {
        printk("Failed to set security (err %d)\n", ret);
    }

    //if (conn->role == BT_CONN_ROLE_CENTRAL) {
    //    printk("central in this connection\n");
    //} else if (conn->role == BT_CONN_ROLE_PERIPHERAL) {
    //    printk("peripheral in this connection\n");
    //}

    btNusSetState(BT_NUS_STATE_CONNECT);
    btNusThreadCreate();
    printk("connected() x\n");
}

static void disconnected(struct bt_conn *conn, uint8_t reason) {
    char addr[BT_ADDR_LE_STR_LEN];

    btNusSetState(BT_NUS_STATE_DISCONNECT);
    btNusThreadAbort();
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

    printk("Disconnected: %s (reason %u)\n", addr, reason);

    if (auth_conn) {
        bt_conn_unref(auth_conn);
        auth_conn = NULL;
    }

    if (_gCurrentConn) {
        bt_conn_unref(_gCurrentConn);
        _gCurrentConn = NULL;
    }
}

#ifdef CONFIG_BT_NUS_SECURITY_ENABLED
static void security_changed(struct bt_conn *conn, bt_security_t level,
                 enum bt_security_err err)
{
    char addr[BT_ADDR_LE_STR_LEN];

    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

    if (!err) {
        printk("Security changed: %s level %u", addr, level);
    } else {
        printk("Security failed: %s level %u err %d", addr,
            level, err);
    }
}

static void auth_passkey_display(struct bt_conn *conn, unsigned int passkey)
{
    char addr[BT_ADDR_LE_STR_LEN];

    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
}

static void auth_cancel(struct bt_conn *conn)
{
    char addr[BT_ADDR_LE_STR_LEN];

    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

    printk("Pairing cancelled: %s", addr);
}


static void pairing_complete(struct bt_conn *conn, bool bonded)
{
    char addr[BT_ADDR_LE_STR_LEN];

    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

    printk("Pairing completed: %s, bonded: %d", addr, bonded);
}


static void pairing_failed(struct bt_conn *conn, enum bt_security_err reason)
{
    char addr[BT_ADDR_LE_STR_LEN];

    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
    bt_conn_disconnect(conn, BT_HCI_ERR_REMOTE_USER_TERM_CONN);
    printk("Pairing failed conn: %s, reason %d", addr, reason);
}


#endif

BT_CONN_CB_DEFINE(conn_callbacks) = {
    .connected    = connected,
    .disconnected = disconnected,
#ifdef CONFIG_BT_NUS_SECURITY_ENABLED
    .security_changed = security_changed,
#endif
};

#if defined(CONFIG_BT_NUS_SECURITY_ENABLED)
static struct bt_conn_auth_cb conn_auth_callbacks = {
    .passkey_display = auth_passkey_display,
    .passkey_confirm = NULL,
    .cancel = auth_cancel,
};

static struct bt_conn_auth_info_cb conn_auth_info_callbacks = {
    .pairing_complete = pairing_complete,
    .pairing_failed = pairing_failed
};
#else
static struct bt_conn_auth_cb conn_auth_callbacks;
//static struct bt_conn_auth_info_cb conn_auth_info_callbacks;
#endif


static void _btReady(const int err) {
    printk("_btReady(%d) e\n", err);
    if (err) {
        printk("Bluetooth init failed (err %d)\n", err);
        return;
    }

    printk("Bluetooth initialized\n");
    
    const int ret = btNusInit();
    if (ret < 0) {
        printk("btNusInit() failed, err=0x%x\n", ret);
        return;
    }
    
    /* Start advertising */
    btAdvStart();
    printk("_btReady(%d) x\n", err);
}

int btInit(void) {
    printk("btInit() e\n");
    int ret;
    if (IS_ENABLED(CONFIG_BT_NUS_SECURITY_ENABLED)) {
        ret = bt_conn_auth_cb_register(&conn_auth_callbacks);
        if (ret != 0) {
            printk("Failed to register authorization callbacks.\n");
            return ERROR_BT_INITIALIZE_FAIL;
        }

        ret = bt_conn_auth_info_cb_register(&conn_auth_info_callbacks);
        if (ret != 0) {
            printk("Failed to register authorization info callbacks.\n");
            return ERROR_BT_INITIALIZE_FAIL;
        }

        ret = bt_passkey_set(_gPasskey);
        if (ret != 0) {
            printk("Failed to set passkey.\n");
            return ERROR_BT_INITIALIZE_FAIL;
        }
    }

    ret = bt_enable(_btReady);
    if (ret != 0) {
        printk("Bluetooth init failed (err 0x%x)\n", ret);
        return ERROR_BT_INITIALIZE_FAIL;
    }

    printk("btInit() x\n");
    return 0;
}

struct bt_conn* btGetConnectHandle(void) {
    return _gCurrentConn;
}
