/*
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <bluetooth/services/nus.h>
#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/kernel.h>
#include <zephyr/logging/log.h>
#include <zephyr/sys/printk.h>
#include "bool.h"
#include "bt_comm.h"
#include "bt_nus.h"
#include "data.h"
#include "datetime.h"
#include "error.h"
#include "localtime.h"

#define BT_NUS_DEBUG             0
#define BT_NUS_CMD_CAPTURE_LEN   7
#define BT_NUS_CMD_MEASURE_LEN   7
#define BT_NUS_CMD_READ_LEN      4
#define BT_NUS_CMD_SKIN_LEN      4
#define BT_NUS_CMD_STOP_LEN      4
#define BT_NUS_CMD_TIME_LEN      DATE_TIME_BUF_LEN
#define BT_NUS_MAX_DATA_LEN      1024
#define BT_NUS_RX_BUF_SIZE       256
#define BT_NUS_TX_BUF_SIZE       1024
#define BT_NUS_THREAD_PRIORITY   (0u)
#define BT_NUS_THREAD_STACK_SIZE (1024u)
#define BT_NUS_WRITE_TIMEOUT     K_MSEC(150)
#define LOG_MODULE_NAME          btnus
LOG_MODULE_REGISTER(LOG_MODULE_NAME);

void   _btNusThread(void);
static K_SEM_DEFINE(_gBtNusReceiveSem,     0, 1);
static K_SEM_DEFINE(_gBtNusSentSem,        0, 1);
static K_SEM_DEFINE(_gBtNusSentEnabledSem, 0, 1);
static K_THREAD_STACK_DEFINE(_gBtNusThreadStack, BT_NUS_THREAD_STACK_SIZE);

typedef enum {
    BT_NUS_CMD_UNKNOWN = 0, 
    BT_NUS_CMD_TIME,
    BT_NUS_CMD_SKIN,  
    BT_NUS_CMD_CAPTURE,
    BT_NUS_CMD_MEASURE,
    BT_NUS_CMD_STOP, 
    BT_NUS_CMD_READ,
} bt_nus_cmd_t;

static bool            _gBtNusTxCompleted = false;
static bool            _gBtNusTxBufFull   = false;
static bt_nus_state_t  _gBtNusState = BT_NUS_STATE_DISCONNECT;
static k_tid_t         _gBtNusThreadId;
static struct k_thread _gBtNusThread;
static uint8_t         _gBtNusTxBuf[BT_NUS_TX_BUF_SIZE];
static uint16_t        _gBtNusTxBufHeader  = 0;
static uint16_t        _gBtNusTxBufTail    = 0;
static uint32_t        _gBtNusMtu    = 0;
static uint32_t        _gBtNusSentBytes = 0;

static void _btNusReceiveCb(struct bt_conn *conn, const uint8_t* const pData, const uint16_t len);
static void _btNusSentCb(struct bt_conn *conn);
static void _btNusSentEnabledCb(enum bt_nus_send_status status);

static struct bt_nus_cb _gNusCb = {
    .received     = _btNusReceiveCb,
    .sent         = _btNusSentCb,
    .send_enabled = _btNusSentEnabledCb
};

static bt_nus_cmd_t _btNusGetCmd( const uint8_t* const pData, const uint16_t len) {
    bt_nus_cmd_t cmd = BT_NUS_CMD_UNKNOWN;
    if ((len == BT_NUS_CMD_CAPTURE_LEN) &&
        (pData[0] == 0x43) && //  "CAPTURE"
        (pData[1] == 0x41) &&
        (pData[2] == 0x50) &&
        (pData[3] == 0x54) &&
        (pData[4] == 0x55) &&
        (pData[5] == 0x52) &&
        (pData[6] == 0x45)) {
        cmd = BT_NUS_CMD_CAPTURE;
    } else if ((len == BT_NUS_CMD_STOP_LEN) &&
        (pData[0] == 0x53) && //  "STOP"
        (pData[1] == 0x54) &&
        (pData[2] == 0x4F) &&
        (pData[3] == 0x50)) {
        cmd = BT_NUS_CMD_STOP;
    } else if ((len == BT_NUS_CMD_TIME_LEN) &&
        (pData[0] == 0x54) && //  "TIME"
        (pData[1] == 0x49) &&
        (pData[2] == 0x4D) &&
        (pData[3] == 0x45)) {
        cmd = BT_NUS_CMD_TIME;
    }

    return cmd;
}

bool _btNusIsTxBufEmpty(void) {
    if (!_gBtNusTxBufFull && _gBtNusTxBufHeader == _gBtNusTxBufTail) {
        return true;
    }

    return false;
}


void _btNusTxBufClear(void) {
    _gBtNusTxBufHeader = 0;
    _gBtNusTxBufTail   = 0;
    _gBtNusTxBufFull   = false;
}

bool _btNusTxBufEmpty(void) {
    if ((!_gBtNusTxBufFull) && (_gBtNusTxBufHeader == _gBtNusTxBufTail)) {
        return true;
    }
    
    return false;
}

uint16_t _btNusTxBufRead(uint8_t* buf, const int len) {
    int i;
    int count = 0;
    
    for(i=0;i<len;i++) {
        if (_gBtNusTxBufFull || _gBtNusTxBufHeader != _gBtNusTxBufTail) {
            buf[i] = _gBtNusTxBuf[_gBtNusTxBufTail];
            _gBtNusTxBufTail++;
            if (_gBtNusTxBufTail >= BT_NUS_TX_BUF_SIZE) {
                _gBtNusTxBufTail = 0;
            }
            _gBtNusTxBufFull = false;
            count++;
        } else {
            break;
        }
    }
    
    return count;
}

uint16_t _btNusTxBufWrite(const uint8_t *buff, const int len) {
    int i;
    uint16_t count = 0;
    if (!_gBtNusTxBufFull) {
        for(i=0;i<len;i++) {
            _gBtNusTxBuf[_gBtNusTxBufHeader] = buff[i];
            _gBtNusTxBufHeader++;
            if (_gBtNusTxBufHeader >= BT_NUS_TX_BUF_SIZE) {
                _gBtNusTxBufHeader = 0;
            }

            count++;
            if (_gBtNusTxBufHeader == _gBtNusTxBufTail) {
                printk("_btNusWriteTxBuf Full\n");
                _gBtNusTxBufFull = true;
                break;
            }
        }
    }
    
    return count;
}

uint32_t _btNusSend(void) {
    //printk("_btNusSend() e\n");
    static uint8_t sbuf[BT_NUS_MAX_DATA_LEN];
    uint16_t len;
    int ret = 0;

    len = 128;

    //printk("len=%u\n", len);
    if (len > 0) {
#if NUS_PPG_DEBUG_ENABLED && AFE4900_TEST_ENABLED
        static uint32_t lastG = 0xFFFFFFFF;
        uint8_t idxA[32];
        const int cnt = dataFindPpgHeaderIndex(len, sbuf, idxA);

        for (int i=0; i<cnt; i++) {
            const uint32_t u32 = ppgData2Uint32(&sbuf[2+idxA[i]]);
            if ((u32 != (lastG+PPG_ONE_PACKET_LENGTH)) && (lastG != 0xFFFFFFFF)) {
                printk("lastG=%u, u32=%u", lastG, u32);
            } else {
                lastG = u32;
            }    
        }
#endif 

#if NUS_GSENSOR_DEBUG_ENABLED && GSENSOR_TEST_ENABLED
        static uint32_t lastX   = 0xFFFFFFFF;
        static uint32_t lastU32 = 0xFFFFFFFF;
        uint8_t idxA[32];
        const int cnt = dataFindGsensorHeaderIndex(len, sbuf, idxA);
        
        for (int i=0; i<cnt; i++) {
            const uint32_t u32 = gsensorData2Uint32(&sbuf[2+idxA[i]]);
            if ((u32 != (lastX+GSENSOR_ONE_PACKET_LENGTH)) && (lastX != 0xFFFFFFFF)) {
                printk("lastX=%u, u32=%u", lastX, u32);
                if (u32 == (lastU32+GSENSOR_ONE_PACKET_LENGTH)) {
                    lastX = u32;
                }
            } else {
                lastX = u32;
            }
            lastU32 = u32;
        }
#endif 

        _gBtNusSentBytes += len;
#if BT_NUS_DEBUG
        printk("bt_nus_send(%p, %p, %u)\n", btGetConnectHandle(), sbuf, len);
#endif
        ret = bt_nus_send(btGetConnectHandle(), sbuf, len);
        if (ret < 0) {
            printk("bt_nus_send() failed, 0x%x\n", ret);
            return ret;
        }
        
        ret = k_sem_take(&_gBtNusSentSem, BT_NUS_WRITE_TIMEOUT);
        if (ret < 0) {
            LOG_WRN("NUS send timeout");
            return ret;
        }
    }

    return ret;
}

static void _btNusReceiveCb(struct bt_conn *conn, const uint8_t* const pData, const uint16_t len) {
    printk("_btReceiveCb(), le=%u e\n", len);
    k_sem_give(&_gBtNusReceiveSem);
    dataPrintChar("nusr", len, pData);
    //dataPrintHex("nusr", len, pData);
    const bt_nus_cmd_t cmd = _btNusGetCmd(pData, len); 
     if (cmd == BT_NUS_CMD_CAPTURE) {
        printk("CAPTURE\n");
        btNusSetState(BT_NUS_STATE_CAPTURE);
    } else if (cmd == BT_NUS_CMD_STOP) {
        printk("STOP\n");
        btNusSetState(BT_NUS_STATE_STOP_BEGIN);
    }  else if (cmd == BT_NUS_CMD_TIME) {
            printk("TIME\n");
            btNusSetState(BT_NUS_STATE_SYNC);
            datetime_t dateTime;
            data2Datetime(&pData[0], &dateTime);
            datetimePrint("datetime sync", &dateTime);
            ltSetDatetime(&dateTime);
    } 

}

static void _btNusSentCb(struct bt_conn *conn) {
#if BT_NUS_DEBUG
    printk("_btNusSentCb(%p)\n", conn);
#endif
    k_sem_give(&_gBtNusSentSem);
}

static void _btNusSentEnabledCb(enum bt_nus_send_status status) {
    if (status == BT_NUS_SEND_STATUS_ENABLED) {
        LOG_DBG("NUS notification has been enabled");
        k_sem_give(&_gBtNusSentEnabledSem);
    }
}


void _btNusThread(void) {
    uint32_t errCode;
    k_sem_take(&_gBtNusSentEnabledSem, K_FOREVER);
    _gBtNusMtu = bt_nus_get_mtu(btGetConnectHandle());
    printk("_gBtNusMtu=%u\n", _gBtNusMtu);

    printk("BT NUS thread starts running\n");
    while(1) {
        const bt_nus_state_t state = btNusGetState(); 
        errCode = _btNusSend();
        if (errCode < 0) {
            printk("Sending NUS data failed\n");
        }
        k_usleep(1);

        if ((errCode < 0) || _btNusTxBufEmpty()) {
             _btNusTxBufClear();
             _gBtNusTxCompleted = true;
        }
        if (state == BT_NUS_STATE_DISCONNECT) {
            btNusThreadAbort();
        }
    } 
}

void btNusThreadAbort(void) {
    printk("btNusThreadAbort() e\n");
    k_thread_abort(_gBtNusThreadId);
}

void btNusThreadCreate(void) {
    printk("btNusThreadCreate() e\n");
    _gBtNusThreadId = k_thread_create(&_gBtNusThread,
                                      _gBtNusThreadStack,
                                      BT_NUS_THREAD_STACK_SIZE,
                                      (k_thread_entry_t) _btNusThread,
                                      NULL,
                                      NULL,
                                      NULL,
                                      BT_NUS_THREAD_PRIORITY,
                                      0,
                                      K_NO_WAIT);
}

int btNusInit(void) {
    const int ret = bt_nus_init(&_gNusCb);
    if (ret < 0) {
        printk("bt_nus_init() failed, err=0x%x\n", ret);
        return ERROR_NUS_INITIALIZE_FAIL;
    }

    return 0;
}

bt_nus_state_t btNusGetState(void) {
    return _gBtNusState;
}

void btNusSetState(const bt_nus_state_t state) {
    _gBtNusState = state;
    btNusPrintState();
}

void btNusThreadResume(void) {
    printk("btNusThreadResume() e\n");
    k_thread_resume(_gBtNusThreadId);
}

void btNusThreadSuspend(void) {
    printk("btNusThreadSuspend() e\n");
    k_thread_suspend(_gBtNusThreadId);
}

uint16_t btNusWriteData(uint8_t *pValue, uint16_t len) {
    const uint16_t ret = _btNusTxBufWrite(pValue, len);
    _gBtNusTxCompleted = true;

    return ret;
}

void btNusPrintState(void) {
    const bt_nus_state_t state = btNusGetState();
    switch (state) {
        case BT_NUS_STATE_DISCONNECT:
        {
            printk("state = disconnect\n"); 
        }
        break;

        case BT_NUS_STATE_CONNECT:
        {
            printk("state = connect\n"); 
        }
        break;
        
        case BT_NUS_STATE_SYNC:
        {
            printk("state = sync\n"); 
        }
        break;
        
        case BT_NUS_STATE_SKIN:
        {
            printk("state = skin\n"); 
        }
        break;
        
        case BT_NUS_STATE_CAPTURE:
        {
            printk("state = capture\n"); 
        }
        break;
        
        case BT_NUS_STATE_MEASURE_BEGIN:
        {
            printk("state = measure begin\n"); 
        }
        break;
        
        case BT_NUS_STATE_MEASURE:
        {
            printk("state = measure\n");
        }
        break;
        
        case BT_NUS_STATE_MEASURE_ENDING:
        {
            printk("state = measure ending\n");
        }
        break;
        
        case BT_NUS_STATE_MEASURE_ENDED:
        {
            printk("state = measure ended\n");
        }
        break;
        
        case BT_NUS_STATE_STOP_BEGIN:
        {
            printk("state = stop begin\n");
        }
        break;
        
        case BT_NUS_STATE_STOP:
        {
            printk("state = stop\n");
        }
        break;
        
        case BT_NUS_STATE_READ_BEGIN:
        {
            printk("state = read begin\n");
        }
        break;
        
        case BT_NUS_STATE_READ:
        {
            printk("state = read\n");
        }
        break;
    }    
}
